'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

exports.RequestSender = RequestSender;

var _isFunction = require('evojs/lib/isFunction');

var _isFunction2 = _interopRequireDefault(_isFunction);

var _isObject = require('evojs/lib/isObject');

var _isObject2 = _interopRequireDefault(_isObject);

var _invariant = require('evojs/lib/invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _checkResultFormat = require('./checkResultFormat');

var _checkResultFormat2 = _interopRequireDefault(_checkResultFormat);

var _interpretResult = require('./interpretResult');

var _interpretResult2 = _interopRequireDefault(_interpretResult);

var _transformers = require('./transformers');

var _Result = require('./Result');

var _Result2 = _interopRequireDefault(_Result);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var defaultNetworkInfo = {
  isConnected: function isConnected() {
    return true;
  }
};

/**
 * @class RequestSender
 * @description
 * HttpRequestSenderFactory allow to send HTTP requests without caring about the implementation by providing
 * a simple interface for both requests and response.
 *
 * An HTTP implementation do the actual work. It must implement the HttpRequestImplement interface.
 */
function RequestSender(initialImplementation) {
  var networkInfo = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : defaultNetworkInfo;

  var implementation = (0, _isFunction2.default)(initialImplementation) ? initialImplementation() : _extends({}, initialImplementation);
  var enableResponseChecking = true;
  var isConnected = networkInfo.isConnected;

  (0, _invariant2.default)((typeof implementation === 'undefined' ? 'undefined' : _typeof(implementation)) === "object" && implementation !== null, "You must provide a valid implementation.");

  if ((0, _isFunction2.default)(implementation.onResult)) {
    implementation.onResult = implementation.onResult.bind(this);
  }

  this.implementation = implementation;
  this.isDisconnected = function () {
    return (0, _isFunction2.default)(isConnected) && isConnected() === false;
  };

  /**
   * Extract
   * @param config
   * @returns {*}
   */
  this.applyTransformersToRequest = function (config) {
    if ((0, _isFunction2.default)(config.transformer) && config.request) {
      // The implementation may have a hook on transforming the request
      // It allows the function to do something on transforming before passing the parameters to the sender
      // Otherwise we just call the transformer
      if ((0, _isFunction2.default)(implementation.onTransformRequestBody)) {
        return implementation.onTransformRequestBody(config.transformer, config.request);
      } else {
        return config.transformer(config.request);
      }
    }

    return config.request;
  };

  /**
   * Create a valid configuration from the request configuration
   * @param config
   * @returns {RequestConfiguration}
   */
  this.createConfiguration = function createConfiguration(config) {
    var request = this.applyTransformersToRequest(config);

    return _extends({
      query: config.query || null,
      request: request || null,
      headers: config.headers || null,
      timeout: config.timeout || null,
      url: config.url,
      method: config.method
    }, config.payload ? { payload: config.payload } : {});
  };

  /**
   * Perform the request
   * @param config
   * @returns {ResultType}
   */
  this.send = function (config) {
    if (this.isDisconnected()) {
      // We trust the connectionStatus to avoid useless request if we are offline
      return new Promise(function (accept) {
        return accept({
          status: "timeout",
          response: null
        });
      });
    }

    config = this.createConfiguration(config);
    if ((0, _isFunction2.default)(implementation.beforeSendingRequest)) {
      config = implementation.beforeSendingRequest(config);
    }

    return implementation.sendRequest(config).then(function (result) {
      if (enableResponseChecking) {
        (0, _checkResultFormat2.default)(result);
      }

      if ((0, _isObject2.default)(implementation) && (0, _isFunction2.default)(implementation.onResult)) {
        implementation.onResult(config, result);
      }

      return new _Result2.default(result);
    }.bind(this));
  };
}

RequestSender.interpretResult = _interpretResult2.default;
RequestSender.Transformer = {
  URITransformer: _transformers.URITransformer,
  FormDataTransformer: _transformers.FormDataTransformer
};

exports.default = {
  /**
   * @param createImplementation
   * @param {NetworkInfoType} networkInfo
   * @returns {RequestSender}
   */
  create: function create(createImplementation) {
    var networkInfo = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : defaultNetworkInfo;

    return new RequestSender(createImplementation, networkInfo);
  }
};