// @flow
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

import invariant from 'invariant';

import { createStore, combineReducers, compose } from 'redux';
import { AsyncStorage } from 'react-native';

import { reduxBatch } from '../util/reduxBatch';
import { persistStore, autoRehydrate } from 'redux-persist';

export type ReducerType = (state : ?mixed, action : { type: mixed }) => mixed;

function StoreManager(){
  this._store = null;
  this._isConfigured = false;
}

export type StoreConfiguration = {
  reducers?: {[key: string]: ReducerType },
  debounce: number,
};

const defaultInitConfiguration : StoreConfiguration = {
  debounce: 500,
};

/**
 * Creates the store and rehydrates it
 *
 * @param config
 * @returns {Promise}
 */
StoreManager.prototype.init = function(config : StoreConfiguration){
  config = Object.assign({}, defaultInitConfiguration, config);

  this._store = createStore(
    combineReducers(config.reducers),
    compose(
      reduxBatch,
      autoRehydrate(),
    ),
  );

  return new Promise((accept) => {
    persistStore(this._store, {
      debounce: config.debounce,
      storage: AsyncStorage,
    }, () => {
      this._isConfigured = true;
      accept(this._store);
    });
  });
};

/**
 * Throws an error if the store is not configured
 */
StoreManager.prototype.assertStoreIsConfigured = function(){
  invariant(
    this.isConfigured() === true,
    "You cannot access any property of the store as it is not configured yet. " +
    "Please call the init function to configure the store."
  )
};

/**
 * Return the state
 * @returns {S}
 */
StoreManager.prototype.getState = function(){
  this.assertStoreIsConfigured();
  return this._store.getState();
};

/**
 *
 * @returns {*<S>}
 */
StoreManager.prototype.dispatch = function(action){
  this.assertStoreIsConfigured();
  return this._store.dispatch(action);
};

/**
 * Returns true if the store is configured
 * @returns {boolean}
 */
StoreManager.prototype.isConfigured = function() : boolean {
  return this._isConfigured;
};

/**
 * Return the store
 * @returns {Store<S>|*|null}
 */
StoreManager.prototype.getStore = function(){
  return this._store;
};

export default {
  create(){
    return new StoreManager();
  }
}