// @flow
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */

import isFunction from 'evojs/lib/isFunction';
import isObject from 'evojs/lib/isObject';
import invariant from 'evojs/lib/invariant';

import checkResultFormat from './checkResultFormat';
import interpretResult from './interpretResult';
import { URITransformer, FormDataTransformer } from './transformers';
import Result from './Result';

import type {
  RequestConfiguration,
  ResultType,
  NetworkInfoType,
} from './definitionTypes';

export type HttpRequestImplementation = {
  sendRequest: (request: RequestConfiguration) => Promise<ResultType>,
  onTransformRequestBody?: <$Request, $Out>(transformer : (request : $Request) => $Out, request : $Request) => ?$Out,
  beforeSendingRequest?: <$Request>(request : $Request) => ?$Request,
  onResult?: (request : RequestConfiguration, result : ResultType) => void,
};

const defaultNetworkInfo : NetworkInfoType = {
  isConnected(){
    return true;
  }
};

/**
 * @class RequestSender
 * @description
 * HttpRequestSenderFactory allow to send HTTP requests without caring about the implementation by providing
 * a simple interface for both requests and response.
 *
 * An HTTP implementation do the actual work. It must implement the HttpRequestImplement interface.
 */
export function RequestSender(
  initialImplementation: () => HttpRequestImplementation | HttpRequestImplementation,
  networkInfo : NetworkInfoType = defaultNetworkInfo
){
  let implementation : HttpRequestImplementation  = isFunction(initialImplementation) ? initialImplementation() : { ...initialImplementation};
  let enableResponseChecking                      = true;
  let isConnected                                 = networkInfo.isConnected;

  invariant(
    typeof implementation === "object" && implementation !== null,
    "You must provide a valid implementation."
  );

  if(isFunction(implementation.onResult)){
    implementation.onResult = implementation.onResult.bind(this);
  }

  this.implementation = implementation;
  this.isDisconnected = () => isFunction(isConnected) && isConnected() === false;

  /**
   * Extract
   * @param config
   * @returns {*}
   */
  this.applyTransformersToRequest = function(config : RequestConfiguration) : ?Object {
    if(isFunction(config.transformer) && config.request){
      // The implementation may have a hook on transforming the request
      // It allows the function to do something on transforming before passing the parameters to the sender
      // Otherwise we just call the transformer
      if(isFunction(implementation.onTransformRequestBody)){
        return implementation.onTransformRequestBody(config.transformer, config.request);
      } else {
        return config.transformer(config.request);
      }
    }

    return config.request;
  };

  /**
   * Create a valid configuration from the request configuration
   * @param config
   * @returns {RequestConfiguration}
   */
  this.createConfiguration = function createConfiguration(config : RequestConfiguration) : RequestConfiguration {
    let request = this.applyTransformersToRequest(config);

    return {
      query: config.query || null,
      request: request || null,
      headers: config.headers || null,
      timeout: config.timeout || null,
      url: config.url,
      method: config.method,
      ...(config.payload ? { payload: config.payload } : {})
    };
  };

  /**
   * Perform the request
   * @param config
   * @returns {ResultType}
   */
  this.send = function(config: RequestConfiguration): Promise<ResultType> {
    if(this.isDisconnected()){
      // We trust the connectionStatus to avoid useless request if we are offline
      return new Promise((accept) => accept({
        status: "timeout",
        response: null,
      }));
    }

    config = this.createConfiguration(config);
    if(isFunction(implementation.beforeSendingRequest)){
      config = implementation.beforeSendingRequest(config);
    }

    return implementation
      .sendRequest(config)
      .then(function(result : ResultType){
        if(enableResponseChecking){
          checkResultFormat(result);
        }

        if(isObject(implementation) && isFunction(implementation.onResult)){
          implementation.onResult(config, result);
        }

        return new Result(result);
      }.bind(this));
  }
}

RequestSender.interpretResult = interpretResult;
RequestSender.Transformer = {
  URITransformer,
  FormDataTransformer,
};

export default {
  /**
   * @param createImplementation
   * @param {NetworkInfoType} networkInfo
   * @returns {RequestSender}
   */
  create(createImplementation: () => HttpRequestImplementation | HttpRequestImplementation, networkInfo : NetworkInfoType = defaultNetworkInfo,){
    return new RequestSender(createImplementation, networkInfo)
  },
}