// @flow
/**
 * Copyright (c) 2017-present, Evosphere.
 * All rights reserved.
 */
import objectMap from 'evojs/lib/objectMap';
import isArray from 'evojs/lib/isArray';
import isFixedNumber from 'evojs/lib/isFixedNumber';
import encodeObjectToURI from 'evojs/lib/encodeObjectToURI';

import type { RequestConfiguration } from '../definitionTypes';
import type { HttpRequestImplementation } from '../RequestSender';

const DEFAULT_TIMEOUT : number = 5000;

/**
 * Return a timeout if any, or by default return 5000
 * @param obj
 * @returns {number}
 */
function getTimeoutDelay(obj : Object) : number {
  if(isFixedNumber(obj.timeout)){
    return obj.timeout;
  }

  return DEFAULT_TIMEOUT;
}

/**
 * Map the headers into a map of headers with scalar values
 * @param headers
 */
function mapHeaders(headers : Object) : Object {
  return objectMap(headers, (value) => {
    if(isArray(value) && value.length === 1){
      return value[0];
    }

    return value;
  })
}

const TimeoutException = {
  name: "TimeoutException",
};

/**
 * make sure the promise rejects once the delay has expired
 * @param ms
 * @param promise
 * @returns {Promise}
 */
export const withTimeout = (ms : number, promise : Promise<*>) : Promise<*> => {
  return new Promise((resolve, reject) => {
    setTimeout(() => {
      reject(TimeoutException);
    }, ms);

    promise.then(resolve, reject);
  });
};

function urlAlreadyHasParameters(url){
  return url.indexOf("?") > -1;
}

function transformURLWithQuery(url, parameters){
  if(urlAlreadyHasParameters(url)){
    return url + "&" + parameters;
  }

  return url + "?" + parameters;
}
/**
 * Creates an adapter around the Fetch API
 * @param fetchFn
 */
const createAdapter = (fetchFn : Function) => () : HttpRequestImplementation => {
  const defaultHeaders = {
    'Content-Type': 'application/x-www-form-urlencoded'
  };

  return {
    sendRequest: (request : RequestConfiguration) => {
      const config : {
        method: string,
        headers: Object,
        body?: string,
      } = {
        method: request.method,
        headers: defaultHeaders,
      };

      if(request.headers){
        config.headers = request.headers;
      }

      // We send POST parameters as x-www-form-urlencoded
      if(request.request && typeof request.request === "object"){
        if(!request.request._parts){
          config.body = encodeObjectToURI(request.request);
        } else {
          // $FlowFixMe
          config.body = request.request;
        }
      }


      // GET parameters must be passed in the URL
      let url = request.url;
      if(request.query){
        url = transformURLWithQuery(request.url, encodeObjectToURI(request.query));
      }

      return withTimeout(getTimeoutDelay(request), fetchFn(url, config))
      .then(response => {
        return response.json()
        .then(data => {
          return {
            status: "done",
            response: {
              statusCode: response.status,
              headers: mapHeaders(response.headers.map),
              data,
            }
          }
        })
        .catch((e) => {
          if(e.name === "SyntaxError"){
            return {
              status: "done",
              response: {
                statusCode: response.status,
                headers: mapHeaders(response.headers.map),
                data: null,
              }
            }
          }
        })
      })
      .catch(error => {
        // See withTimeout
        if(error === TimeoutException){
          return {
            status: "timeout",
            response: null,
          }
        }

        return {
          status: "network-error",
          response: null,
        }
      });
    }
  }
};

export default createAdapter;